# Standard Libraries
from dataclasses import dataclass
from pathlib import Path
from enum import Enum
from typing import Optional

# Dependencies
import rasterio as rio  # type: ignore
from rasterio.crs import CRS  # type: ignore
import numpy.typing as npt
import h5py as h5  # type: ignore

# Relative Imports
from .typing import PathLike, Spectrum
from .endmember import EndMemberGroup, EndMember
from .model_math import UnMixedCube


type GeotransformType = tuple[float, float, float, float, float, float]


class UnsupportedFileExtensionError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


@dataclass
class ModelResult:
    savefile: PathLike
    modelID: str
    unmixed_image: UnMixedCube
    endmembers: EndMemberGroup
    rsquared: npt.NDArray


class SaveMode(Enum):
    HDF5 = ".hdf5"
    SMA = ".sma"
    CSV = ".csv"


def save_model_result(res: ModelResult):
    file_ext = Path(res.savefile).suffix

    try:
        mode = SaveMode(file_ext)
    except ValueError:
        raise UnsupportedFileExtensionError(
            f"{file_ext} is not a supported" "file type."
        )
    match mode:
        case SaveMode.HDF5:
            if Path(res.savefile).is_file():
                open_flag = "r+"
            else:
                open_flag = "w"

            with h5.File(res.savefile, open_flag) as f:
                try:
                    g = f.create_group(res.modelID)
                except ValueError:
                    del f[res.modelID]
                    g = f.create_group(res.modelID)

                g.attrs["wavelengths"] = res.endmembers.endmember_list[
                    0
                ].spectrum.wvl
                gg = g.create_group("endmembers")
                for n, name in enumerate(res.endmembers.endmember_name_list):
                    gg.create_dataset(
                        name,
                        data=res.endmembers.endmember_list[n].spectrum.data,
                    )
                    gg[name].attrs["index"] = n
                g.create_dataset("fractions", data=res.unmixed_image.fracs)
                g.create_dataset("residuals", data=res.unmixed_image.res)
                g.create_dataset("model", data=res.unmixed_image.model)
                g.create_dataset("rsquared", data=res.rsquared)
        case SaveMode.SMA:
            raise NotImplementedError(
                "Saving to .sma has not been implemented."
            )
        case SaveMode.CSV:
            raise NotImplementedError(
                "Saving to .csv has not been implemented."
            )


def load_model_result(p: PathLike, model_name: str):
    with h5.File(p) as f:
        g = f[model_name]
        endmember_list_with_idx = [
            [
                EndMember(
                    name,
                    Spectrum(dat[...], g.attrs["wavelengths"]),  # type: ignore
                ),
                dat.attrs["index"][...],  # type: ignore
            ]
            for (name, dat) in g["endmembers"].items()  # type: ignore
        ]
        endmember_list_sorted = [
            i[0] for i in sorted(endmember_list_with_idx, key=lambda x: x[1])
        ]

        endmember_grp = EndMemberGroup(endmember_list_sorted)

        return ModelResult(
            p,
            model_name,
            UnMixedCube(
                g["model"][...],  # type: ignore
                g["fractions"][:, :, :-1],  # type: ignore
                g["residuals"][...],  # type: ignore
            ),
            endmember_grp,
            g["rsquared"][...],  # type: ignore
        )


def write_model_to_gis(
    model: ModelResult,
    geotransform: GeotransformType,
    crs: str,
    save_directory: str | Path,
    name_prefix: Optional[str] = None,
):
    profile = {
        "driver": "ENVI",
        "height": model.unmixed_image.fracs.shape[0],
        "width": model.unmixed_image.fracs.shape[1],
        "count": model.unmixed_image.fracs.shape[2],
        "transform": geotransform,
        "crs": CRS.from_wkt(crs),
        "dtype": "float32",
        "nodata": -999,
    }

    if name_prefix is not None:
        file_name = f"{name_prefix}_"
    else:
        file_name = f"{model.modelID}_"
    with rio.open(
        Path(save_directory, f"{file_name}fractions").with_suffix(".bsq"),
        "w",
        **profile,
    ) as f:
        for n in range(model.unmixed_image.fracs.shape[-1]):
            f.write(model.unmixed_image.fracs[:, :, n], n + 1)
    with open(
        Path(save_directory, f"{file_name}fractions").with_suffix(".hdr"), "a"
    ) as f:
        f.write("band_names = {\n")
        for i in model.endmembers.endmember_name_list:
            f.write(f"    {i},\n")
        f.write("}")

    profile["count"] = 1
    with rio.open(
        Path(save_directory, f"{file_name}residual").with_suffix(".bsq"),
        "w",
        **profile,
    ) as f:
        f.write(model.rsquared, 1)



# Standard Libraries
from dataclasses import dataclass
from typing_extensions import Annotated
from typing import Tuple

# Dependencies
import numpy as np
import numpy.typing as npt

# Relative Imports
from .typing import ImageCubeLike


@dataclass
class MixedCube:
    """
    Represents an made up of pixels that are each an areal mixture of spectral
    endmembers.

    Attributes
    ----------
    G: NDArray[np.float32, (2,)]
        Least squares design matrix. The array is of shape NxM where N is the
        number of spectral bands and M is the number of endmembers in the
        unmixing model.
    d: NDArray[np.float32, (3,)]
        Data matrix for every pixel. This is just the spectral cube data.
    """

    G: Annotated[npt.NDArray[np.float32], (2,)]
    d: Annotated[npt.NDArray[np.float32], (3,)]


@dataclass
class UnMixedCube:
    """
    The results of spectral unmixing.

    Attributes
    ----------
    model: NDArray[np.float32, (3,)]
        Modeled spectral cube, where each pixel is a linear combination of
        endmembers that best fits the actual data.
    fracs: NDArray[np.float32, (3,)]
        3D array representing the modeled fractions of endmembers at all
        pixels.
    res: NDArray[np.float32, (2,)]
        Pixel-by-pixel model residuals. That is, a Spectrum-Like vector for
        each pixel that represents model - data in each spectral band.
    """

    model: ImageCubeLike
    fracs: ImageCubeLike
    res: ImageCubeLike


def _solve_cube(
    G: Annotated[npt.NDArray[np.float32], (2,)],
    d: Annotated[npt.NDArray[np.float32], (3,)],
) -> Tuple[ImageCubeLike, ImageCubeLike, ImageCubeLike]:
    prefix = np.linalg.inv(G.T @ G) @ G.T

    fracs = np.einsum("ij,...j->...i", prefix, d)

    model = np.einsum("ij,...j->...i", G, fracs)

    res = model - d

    return model, fracs, res


def unmix_spectral_cube(mixed_cube: MixedCube, add_to_one: bool = True):
    G = mixed_cube.G
    d = mixed_cube.d

    if add_to_one:
        G_aug = np.hstack((G, np.ones([G.shape[0], 1])), dtype=np.float32)
        bottom_row = np.append(np.ones(G.shape[1]), 0)
        G_aug = np.vstack((G_aug, bottom_row), dtype=np.float32)

        d_aug = np.concat(
            [d, np.ones(d.shape[:2])[:, :, None]], axis=2, dtype=np.float32
        )

        model, fracs, res = _solve_cube(G_aug, d_aug)
    else:
        model, fracs, res = _solve_cube(G, d)

    return UnMixedCube(model, fracs, res)



from dataclasses import dataclass
from typing import Optional

from PySide6.QtGui import QAction
from PySide6.QtWidgets import QWidget


@dataclass
class ActionSpec:
    text: str
    callback_name: str
    shortcut: Optional[str] = None
    _action: Optional[QAction] = None

    def build(self, parent: QWidget, receiver: object) -> QAction:
        if self._action is None:
            action = QAction(self.text, parent)
            if self.shortcut is not None:
                action.setShortcut(self.shortcut)

            callback = getattr(receiver, self.callback_name)

            action.triggered.connect(callback)
            self._action = action
        return self._action


@dataclass
class ActionCatalog:
    open_frac = ActionSpec(text="Fraction Cube", callback_name="load_frac")
    open_resi = ActionSpec(text="Residual Cube", callback_name="load_resi")
    open_model = ActionSpec(text="Hypmix Model", callback_name="load_model")
    set_base = ActionSpec(text="Set Base", callback_name="set_base")



# Standard Libraries
from dataclasses import dataclass, field

# Dependencies
import numpy as np

# Relative Imports
from .typing import ImageCube, Spectrum, PathLike
from .io import ModelResult
from .model_math import unmix_spectral_cube, MixedCube
from .endmember import EndMember, EndMemberGroup


class EndmemberAlreadyExistsError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


@dataclass
class ModelState:
    has_reflector: bool = False
    has_shade: bool = False
    has_blackbody: bool = False
    endmember_count: int = 0


@dataclass
class MixtureModel:
    """
    Parameters
    ----------
    endmembers: list[EndMember]
        List of model endmembers to use.
    data_cube: ImageCube
        Spectral data cube.
    """

    endmembers: list[EndMember]
    data_cube: ImageCube
    state: ModelState = field(default_factory=ModelState)

    def __post_init__(self):
        if self.data_cube.bands_first:
            self.data_cube.data = np.transpose(self.data_cube.data, (1, 2, 0))
        self.state.endmember_count = len(self.endmembers)

    def add_endmember(self, endmember: EndMember) -> None:
        if endmember.name in [i.name for i in self.endmembers]:
            raise EndmemberAlreadyExistsError(
                f"{endmember} has the same name as an existing endmember."
            )
        self.endmembers.append(endmember)
        self.state.endmember_count += 1

    def add_virtual_reflector(self):
        if not self.state.has_reflector:
            self.endmembers.append(
                EndMember(
                    "Reflector",
                    Spectrum(
                        np.ones(len(self.data_cube.wvl)), self.data_cube.wvl
                    ),
                )
            )
            self.state.endmember_count += 1
            self.state.has_reflector = True

    def add_virtual_shade(self):
        if not self.state.has_shade:
            self.endmembers.append(
                EndMember(
                    "Shade",
                    Spectrum(
                        0.001 * np.ones(len(self.data_cube.wvl)),
                        self.data_cube.wvl,
                    ),
                )
            )
            self.state.endmember_count += 1
            self.state.has_shade = True
        else:
            raise EndmemberAlreadyExistsError(
                "Virtual Shade already exists in this model."
            )

    def add_virtual_blackbody(self) -> None:
        if not self.state.has_blackbody:
            self.endmembers.append(
                EndMember(
                    "Blackbody",
                    Spectrum(
                        np.ones(len(self.data_cube.wvl)), self.data_cube.wvl
                    ),
                )
            )
            self.state.endmember_count += 1
            self.state.has_blackbody = True
        else:
            raise EndmemberAlreadyExistsError(
                "Virtual blackbody already exists in this model."
            )

    def run(self, dst_path: PathLike, modelID: str) -> ModelResult:
        G = np.empty(
            [len(self.data_cube.wvl), len(self.endmembers)], dtype=np.float32
        )
        for n, em in enumerate(self.endmembers):
            G[:, n] = em.spectrum.data
        mixed_cube = MixedCube(G, self.data_cube.data)
        unmixed_cube = unmix_spectral_cube(mixed_cube)
        rsquared = np.sqrt(np.sum(unmixed_cube.res**2, axis=2))

        result = ModelResult(
            dst_path,
            modelID,
            unmixed_cube,
            EndMemberGroup(self.endmembers),
            rsquared,
        )

        return result



# Standard Libraries
from dataclasses import dataclass
from typing import TypeVar
from typing_extensions import Annotated
import os
from pathlib import Path

# Dependencies
import numpy.typing as npt
import numpy as np

PathLike = str | os.PathLike | Path

DType = TypeVar("DType", bound=np.generic)
SpectrumLike = Annotated[npt.NDArray[DType], (1,)]
ImageCubeLike = Annotated[npt.NDArray[DType], (3,)]
ImageLike = Annotated[npt.NDArray[DType], (2,)]


@dataclass
class Spectrum:
    data: SpectrumLike
    wvl: SpectrumLike


@dataclass
class ImageCube:
    data: ImageCubeLike
    wvl: SpectrumLike
    bands_first: bool = False

    def __post_init__(self):
        bands_dim = np.argmin(self.data.shape)
        self.data = np.moveaxis(self.data, bands_dim, -1)



# Built-Ins
from dataclasses import dataclass
from pathlib import Path

# Dependency
from PySide6.QtWidgets import (
    QMainWindow,
    QHBoxLayout,
    QWidget,
    QMenu,
    QFileDialog,
    QDockWidget,
)
from PySide6.QtCore import Qt
import pyqtgraph as pg  # type: ignore
import numpy as np

from hypmix.io import load_model_result

# Local Imports
from .image_view_container import ImViewContainer
from .model_tree import ModelTreeWidget
from .endmember_viewer import EndmemberViewerWidget
from .catalog.actions import ActionCatalog

# Top-Level Imports
from hypmix.file_opening_utils import open_cube

USER_ROLE = Qt.ItemDataRole.UserRole
VERSION_NUM = "0.2.0"


@dataclass
class MasterState:
    base_dir: Path = Path.cwd()


class MixView(QMainWindow):
    def __init__(
        self,
        frac: Path | None = None,
        resi: Path | None = None,
        model: Path | None = None,
        base: Path | None = None,
    ) -> None:
        super().__init__()

        self.state = MasterState()

        self.setWindowTitle("MixView v0.2.0")

        self.frac_view = pg.ImageView()
        self.resi_view = pg.ImageView()
        self.model_tree = ModelTreeWidget()
        self.em_view = EndmemberViewerWidget()
        self.tree_dock: QDockWidget | None = None

        self.frac_container = ImViewContainer(self.frac_view, parent=self)
        self.resi_container = ImViewContainer(
            self.resi_view, parent=self, color_extremes=False
        )

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.frac_container)
        main_layout.addWidget(self.resi_container)

        main_widget = QWidget()
        main_widget.setLayout(main_layout)

        self.setCentralWidget(main_widget)

        # Make the model tree dockable on the right (can be moved/floated)
        tree_dock = QDockWidget("Models", self)
        tree_dock.setWidget(self.model_tree)
        tree_dock.setAllowedAreas(
            Qt.DockWidgetArea.LeftDockWidgetArea
            | Qt.DockWidgetArea.RightDockWidgetArea
        )
        tree_dock.setMinimumWidth(200)
        self.tree_dock = tree_dock

        em_dock = QDockWidget("Endmembers", self)
        em_dock.setWidget(self.em_view)
        em_dock.setAllowedAreas(
            Qt.DockWidgetArea.TopDockWidgetArea
            | Qt.DockWidgetArea.BottomDockWidgetArea
        )
        em_dock.setMinimumHeight(200)
        self.em_dock = em_dock

        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, em_dock)
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, tree_dock)

        menubar = self.menuBar()

        file_menu = QMenu("File")
        data_menu = QMenu("Data")
        menubar.addMenu(file_menu)
        menubar.addMenu(data_menu)

        open_menu = QMenu("Open")
        file_menu.addMenu(open_menu)

        file_menu.addAction(ActionCatalog.set_base.build(main_widget, self))

        open_menu.addAction(ActionCatalog.open_frac.build(main_widget, self))
        open_menu.addAction(ActionCatalog.open_resi.build(main_widget, self))
        open_menu.addAction(ActionCatalog.open_model.build(main_widget, self))

        data_menu.addAction(tree_dock.toggleViewAction())
        data_menu.addAction(em_dock.toggleViewAction())

        self.model_tree.tree.itemSelectionChanged.connect(self.set_model)

        if frac is not None:
            self.load_frac(fp=frac)
        if resi is not None:
            self.load_resi(fp=resi)
        if base is not None:
            self.state.base_dir = base
        if model is not None:
            self.load_model(fp=model)

        self.resize(1400, 1000)

    def load_model(self, *, fp: Path | None = None):
        if fp is None:
            fp_str, fp_type = QFileDialog.getOpenFileName(
                caption="Select Model Result",
                filter=("HDF5 Files (*.hdf5)"),
                dir=str(self.state.base_dir),
            )
            fp = Path(fp_str)

        self.model_tree.add_model(fp)

    def load_frac(self, *, fp: Path | None = None):

        if fp is None:
            fp_str, fp_type = QFileDialog.getOpenFileName(
                caption="Select Fraction Cube",
                filter=(
                    "Rasterio-Compatible Files (*.bsq *.img *.tif);;"
                    "Spectral Cube Files (*.spcub *.geospcub)"
                ),
                dir=str(self.state.base_dir),
            )
            fp = Path(fp_str)
        cube, _suffix = open_cube(fp)
        self.set_frac(cube)

    def load_resi(self, *, fp: Path | None = None):
        if fp is None:
            fp_str, fp_type = QFileDialog.getOpenFileName(
                caption="Select Residual Cube",
                filter=(
                    "Rasterio-Compatible Files (*.bsq *.img *.tif);;"
                    "Spectral Cube Files (*.spcub *.geospcub)"
                ),
                dir=str(self.state.base_dir),
            )
            fp = Path(fp_str)
        cube, _suffix = open_cube(fp)
        self.set_resi(cube)

    def set_frac(self, cube: np.ndarray):
        self.frac_view.setImage(cube, axes={"y": 0, "x": 1, "t": 2})
        self.frac_view.setLevels(0, 1)

    def set_resi(self, cube: np.ndarray):
        self.resi_view.setImage(cube, axes={"y": 0, "x": 1, "t": 2})
        idx = self.resi_view.currentIndex
        vals = cube[np.isfinite(cube[:, :, idx]), idx]
        lo, hi = np.percentile(vals, [0.5, 99.5])
        self.resi_view.setLevels(lo, hi)

    def set_base(self):
        fp = QFileDialog.getExistingDirectory()
        self.state.base_dir = Path(fp)

    def set_model(self) -> None:
        selection = self.model_tree.get_selection_path()
        if selection is None:
            return
        model = load_model_result(selection.fp, selection.model)
        self.set_frac(model.unmixed_image.fracs)
        self.set_resi(model.unmixed_image.res)
        self.frac_container.connect_title(model.endmembers.endmember_name_list)
        wvl = list(model.endmembers.endmember_list[0].spectrum.wvl)
        wvl = [f"{str(i)} nm" for i in wvl]
        self.resi_container.connect_title(wvl)
        self.em_view.show_endmembers(model)



# Standard Libraries
from dataclasses import dataclass
from typing import Tuple

# Dependencies
import numpy as np

# Relative Imports
from .typing import Spectrum, PathLike


@dataclass
class EndMember:
    name: str
    spectrum: Spectrum


@dataclass
class InSceneEndMember(EndMember):
    coord: Tuple[int, int]


@dataclass
class ExternalEndMember(EndMember):
    file_path: PathLike


class EndMemberGroup:
    def __init__(self, endmember_list: list[EndMember]):
        self.endmember_list = endmember_list
        self.endmember_array = np.empty(
            (len(endmember_list[0].spectrum.data), len(endmember_list)),
            dtype=np.float32,
        )
        self.endmember_name_list: list[str] = []

        for n, i in enumerate(endmember_list):
            self.endmember_array[:, n] = i.spectrum.data
            self.endmember_name_list.append(i.name)

    def __len__(self):
        return len(self.endmember_name_list)


def read_endmember_set(
    names: list[str], arr: np.ndarray, wvl: np.ndarray
) -> list[EndMember]:
    em_list: list[EndMember] = []
    for n, name in enumerate(names):
        spec = Spectrum(arr[n, :], wvl)
        em_list.append(EndMember(name, spec))
    return em_list



from PySide6.QtWidgets import QWidget, QHBoxLayout
import pyqtgraph as pg  # type: ignore
import hypmix
import cmap
import reflspeckit as rsk


class EndmemberViewerWidget(QWidget):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)

        self.plot_widget = pg.PlotWidget()
        self.contrem_widget = pg.PlotWidget()
        self.legend = self.plot_widget.addLegend()

        layout = QHBoxLayout()
        layout.addWidget(self.plot_widget)
        layout.addWidget(self.contrem_widget)
        self.setLayout(layout)

    def show_endmembers(self, model: hypmix.ModelResult):
        item1 = self.plot_widget.getPlotItem()
        item2 = self.contrem_widget.getPlotItem()
        if item1 is not None:
            item1.clear()
        if item2 is not None:
            item2.clear()
        base_cmap = cmap.Colormap("crameri:hawaii")
        ncolors = len(model.endmembers.endmember_list)
        cmap_lut = base_cmap.lut(ncolors) * 255
        for n, em in enumerate(model.endmembers.endmember_list):
            c = pg.mkColor(tuple(cmap_lut[n, :]))
            _spec = pg.PlotDataItem(
                em.spectrum.wvl,
                em.spectrum.data,
                name=em.name,
                pen=pg.mkPen(color=c, width=1),
            )

            rspec = rsk.Spec1D(em.spectrum.data, em.spectrum.wvl)
            rspec.outlier_removal()
            rspec.noise_reduction("box_filter", 7)
            rspec.continuum_removal("double_line")

            _contrem = pg.PlotDataItem(
                rspec.wavelength.values,
                rspec.contrem,
                name=em.name,
                pen=pg.mkPen(color=c, width=1),
            )

            self.plot_widget.addItem(_spec)
            self.contrem_widget.addItem(_contrem)



# Built-Ins
from dataclasses import dataclass
from pathlib import Path

# Dependencies
from PySide6.QtWidgets import (
    QTreeWidget,
    QTreeWidgetItem,
    QWidget,
    QVBoxLayout,
)
from PySide6.QtCore import Qt
import h5py as h5  # type: ignore

USER_ROLE = Qt.ItemDataRole.UserRole


@dataclass
class LoadedModel:
    header: str
    fp: Path
    model: str = ""


class ModelTreeWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.tree = QTreeWidget(self)
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)
        self.setMinimumWidth(200)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.tree)
        self.setLayout(layout)

    def add_model(self, fp: Path):
        with h5.File(fp) as f:
            top_level_groups = [
                name for name, obj in f.items() if isinstance(obj, h5.Group)
            ]
        self._populate_tree(fp.stem, top_level_groups, fp)

    def _populate_tree(self, heading: str, values: list[str], fp: Path):
        parent_item = QTreeWidgetItem([heading])
        parent_item.setFlags(
            parent_item.flags() & ~Qt.ItemFlag.ItemIsSelectable
        )
        self.tree.addTopLevelItem(parent_item)

        for child_text in values:
            item = QTreeWidgetItem(parent_item, [child_text])
            metadata = LoadedModel(header=heading, fp=fp, model=child_text)
            item.setData(0, USER_ROLE, metadata)

        parent_item.setExpanded(True)

    def get_selection_path(self) -> LoadedModel | None:
        items = self.tree.selectedItems()
        if not items:
            return None

        return items[0].data(0, USER_ROLE)



from .endmember import InSceneEndMember, ExternalEndMember, EndMember
from .io import ModelResult, save_model_result, load_model_result
from .run_model import MixtureModel
from .typing import Spectrum
from .helper_functions import open_mixview

__all__ = [
    "InSceneEndMember",
    "ExternalEndMember",
    "EndMember",
    "ModelResult",
    "save_model_result",
    "load_model_result",
    "MixtureModel",
    "Spectrum",
    "open_mixview",
]



from PySide6.QtWidgets import QApplication
from hypmix.mixview.main_window import MixView
import sys


def open_mixview():
    app = QApplication(sys.argv)
    main = MixView()
    main.show()
    app.exec()


if __name__ == "__main__":
    open_mixview()



# Dependencies
from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel
import pyqtgraph as pg  # type: ignore
import cmap
import numpy as np


class ImViewContainer(QWidget):
    def __init__(
        self,
        imview_widget: pg.ImageView,
        parent: QWidget | None = None,
        color_extremes: bool = True,
    ) -> None:
        super().__init__(parent)

        layout = QVBoxLayout()
        self.title = QLabel()
        self.imview_widget = imview_widget

        custom_cmap = cmap.Colormap("matlab:gray")
        lut = (custom_cmap.lut(N=8) * 255).astype(np.uint8)
        hist_item = self.imview_widget.getHistogramWidget().item
        if color_extremes:
            pos = np.linspace(0.01, 0.99, lut.shape[0] - 2)
            lut[0] = [255, 0, 0, 255]  # Red
            lut[-1] = [0, 255, 255, 255]  # Cyan
            pos = np.insert(pos, 0, 0)
            pos = np.append(pos, 1)
        else:
            pos = np.linspace(0.0, 1.0, lut.shape[0])

        lut_cmap = pg.ColorMap(pos, lut)
        hist_item.gradient.setColorMap(lut_cmap)

        layout.addWidget(self.title)
        layout.addWidget(self.imview_widget)

        self.setLayout(layout)

    def connect_title(self, lbls: list):
        def _update_title():
            title = lbls[self.imview_widget.currentIndex]
            self.title.setText(title)

        _update_title()
        self.imview_widget.timeLine.sigPositionChanged.connect(_update_title)



# flake8: noqa

from PySide6.QtWidgets import QApplication
from hypmix.mixview.main_window import MixView
import sys
from pathlib import Path


def mixview():
    app = QApplication(sys.argv)

    # For dev
    # main = MixView(
    #     base=Path("D:/moon_data/m3/Gruithuisen_Region/Gruithuisen_Mosaics/"),
    #     model=Path(
    #         "D:/moon_data/m3/Gruithuisen_Region/Gruithuisen_Mosaics/M3G_GDOMES_mixmodel.hdf5"
    #     ),
    # )

    main = MixView()
    main.show()
    app.exec()


def main():
    mixview()



"""
File Opening Utilities

This module provides a centralized, type-safe dispatch for opening three
categories of files for the operation ofthe SpecView GUI.

Supported File Types
--------------------


Available Helper Functions::

    from pathlib import Path

    wvl = open_wvl(Path("wvlvals.wvl"))
    img = open_cube(Path("cube.geospcub"))

"""

# Built-Ins
from pathlib import Path
from typing import Protocol
from dataclasses import dataclass

# Dependencies
import numpy as np
import rasterio as rio  # type: ignore
import spectralio as sio
import re


# ---- Handling Wavelength Data Files ----


class WvlHandler(Protocol):
    """
    Protocol for handling wavelength (or other context data) files.
    """

    def __call__(self, path: Path) -> np.ndarray: ...

    """
    Handle a file at the given path.

    Parameters
    ----------
    path: Path
        Path to an existing file whose suffix is one of the following

        - .wvl
        - .hdr
        - .txt
        - .csv

    Returns
    -------
    wvl_array: np.ndarray
        A 1D numpy array that holds the wavelength values.

    Raises
    ------
    OSError
        If the file cannot be read.
    """


def open_wvl_file(path: Path) -> np.ndarray:
    """Reads .wvl files using `spectralio`"""
    wvl = sio.read_wvl(path)
    return wvl.asarray()


def open_hdr_file(path: Path) -> np.ndarray:
    """Read an ENVI .hdr file"""
    wvl_pattern = re.compile(r"wavelength\s=\s\{([^}]*)\}")
    with open(path, "r") as f:
        file_contents = f.read()
    result = re.findall(wvl_pattern, file_contents)
    if len(result) == 0:
        raise OSError("Unable to open .hdr file. Is there a wavelength field?")
    vals = np.asarray([float(i) for i in result[0].split(",")])
    return vals


def open_txt_file(path: Path) -> np.ndarray:
    """
    Read a .txt file. Wavelength values should be seperated by commas.
    """
    with open(path, "r") as f:
        contents = f.read()
    vals = contents.split(",")
    if vals[-1] == " ":
        vals = vals[:-1]
    return np.asarray([float(i) for i in vals])


def open_csv_file(path: Path) -> np.ndarray:
    """
    Opens a csv file where there is one row of headers and at least one is
    "wavelength". Make sure there are no spaces around the commas!
    """
    arr = np.loadtxt(path, delimiter=",", dtype=str)
    idx = np.argwhere(np.char.lower(arr[0, :]) == "wavelength")
    if idx.size == 0:
        raise ValueError("No Wavelength column exists.")
    if idx.size > 1:
        raise ValueError("Multiple Wavelength columns exist.")
    return arr[1:, idx[0][0]].astype(float)


# Mapping from lowercase string file extensions to handler functions.
WVL_HANDLERS: dict[str, WvlHandler] = {
    ".wvl": open_wvl_file,
    ".hdr": open_hdr_file,
    ".txt": open_txt_file,
    ".csv": open_csv_file,
}


def open_wvl(path: str | Path) -> np.ndarray:
    """
    Open a file that stores wavelength information.

    This function inspects the files extension name and passes it to the
    appropriate handler for that file type.

    Parameters
    ----------
    path: str or Path
        Path to file containing wavelength data that is to be opened.

    Raises
    ------
    FileNotFoundError
        If the file does not exist.
    ValueError
        If the file does not have a valid extension.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(path)

    suffix = path.suffix.lower()

    handler = WVL_HANDLERS.get(suffix)

    if handler is None:
        raise ValueError(f"Unsupported file type: {suffix}")

    return handler(path)


# ---- Handling Cube Data Files ----
@dataclass
class CubeAxisOrder:
    """
    Holds information for orienting spectral data cubes.

    Parameters
    ----------
    x: int
        Axis index for the horizontal image direction.
    y: int
        Axis index for the vertical image direction.
    b: int
        Axis index for the spectral band (or other context data) direction.
    """

    x: int
    y: int
    b: int


class CubeHandler(Protocol):
    """
    Protocol for handling cube data files.
    """

    def __call__(self, path: Path, axis_map: dict[str, int]) -> np.ndarray: ...

    """
    Handle a file at the given path.

    Parameters
    ----------
    path: Path
        Path to an existing file whose suffix is one of the following

        - .spcub
        - .geospcub
        - .bsq
        - .img
        - .tif



    Returns
    -------
    cube_array: np.ndarray
        A 3D numpy array where axis 0 is the vertical image dimension, axis 1
        is the horizontal image dimension and axis 2 is the wavelength (or
        other context data) dimension.

    Raises
    ------
    OSError
        If the file cannot be read.
    """


def open_spcub_cube(path: Path, axis_map: dict[str, int]) -> np.ndarray:
    """
    Read .spcub or .geospcub files using `spectralio`
    """
    cub_obj: sio.Spectrum3D
    if path.suffix.lower() == ".geospcub":
        cub_obj = sio.read_spec3D(path, kind="geospcub")
    elif path.suffix.lower() == ".spcub":
        cub_obj = sio.read_spec3D(path, kind="spcub")
    else:
        raise ValueError(
            "Invalid File Type passed to `open_spcub_file()`: "
            f"{path.suffix.lower()}"
        )
    return cub_obj.load_raster(bbl=True)


def open_rasterio_cube(path: Path, axis_map: dict[str, int]) -> np.ndarray:
    """
    Reads any rasterio-compatible file type.
    """
    try:
        axis_order_obj = CubeAxisOrder(**axis_map)
    except TypeError:
        raise TypeError(
            "Invalid axis_order dictionary with keys: "
            f"{list(axis_map.keys())}. The keys should be ['x', 'y', 'b']"
            " for the horizontal, vertical and spectral (or other) dimension,"
            " respectively."
        )

    with rio.open(path, "r") as f:
        cube_array = f.read()
    transpose_order = (axis_order_obj.y, axis_order_obj.x, axis_order_obj.b)
    cube_array = np.transpose(cube_array, transpose_order)
    return cube_array


# Mapping from lowercase file extension to handler function.
CUBE_HANDLERS: dict[str, CubeHandler] = {
    ".spcub": open_spcub_cube,
    ".geospcub": open_spcub_cube,
    ".bsq": open_rasterio_cube,
    ".img": open_rasterio_cube,
    ".tif": open_rasterio_cube,
}


def open_cube(
    path: str | Path, axis_map: dict[str, int] = {"x": 2, "y": 1, "b": 0}
) -> tuple[np.ndarray, str]:
    """
    Open a file that stores spectral (or other) cube-based information.

    This function inspects the files extension name and passes it to the
    appropriate handler for that file type.

    Parameters
    ----------
    path: str or Path
        Path to file containing wavelength data that is to be opened.

    Returns
    -------
    cube_array: np.ndarray
        3D Array of spectral cube data.
    file_suffix: str
        Lowercase suffix of the file name.

    Raises
    ------
    FileNotFoundError
        If the file does not exist.
    ValueError
        If the file does not have a valid extension.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(path)

    suffix = path.suffix.lower()

    handler = CUBE_HANDLERS.get(suffix)

    if handler is None:
        raise ValueError(f"Unsupported file type: {suffix}")

    return (handler(path, axis_map), suffix)
